@@ -1,5 +1,6 @@
 package org.bouncycastle.asn1;
 
+import java.io.EOFException;
 import java.io.IOException;
 import java.io.InputStream;
 
@@ -28,37 +29,67 @@ class IndefiniteLengthInputStream
         _eofOn00 = eofOn00;
     }
 
-    void checkForEof()
-        throws IOException
+    boolean checkForEof()
     {
         if (_eofOn00 && (_b1 == 0x00 && _b2 == 0x00))
         {
             _eofReached = true;
             setParentEofDetect(true);
         }
+        return _eofReached;
     }
 
-    public int read()
+    public int read(byte[] b, int off, int len)
         throws IOException
     {
-        checkForEof();
+        // Only use this optimisation if we aren't checking for 00
+        if (_eofOn00 || len < 3)
+        {
+            return super.read(b, off, len);
+        }
 
         if (_eofReached)
         {
             return -1;
         }
 
+        int numRead = _in.read(b, off + 2, len - 2);
+
+        if (numRead < 0)
+        {
+            // Corrupted stream
+            throw new EOFException();
+        }
+
+        b[off] = (byte)_b1;
+        b[off + 1] = (byte)_b2;
+
+        _b1 = _in.read();
+        _b2 = _in.read();
+
+        if (_b2 < 0)
+        {
+            // Corrupted stream
+            throw new EOFException();
+        }
+
+        return numRead + 2;
+    }
+
+    public int read()
+        throws IOException
+    {
+        if (checkForEof())
+        {
+            return -1;
+        }
+
         int b = _in.read();
 
-        //
-        // strictly speaking we should return b1 and b2, but if this happens the stream
-        // is corrupted so we are already in trouble.
-        //
         if (b < 0)
         {
-            _eofReached = true;
-
-            return -1;
+            // Corrupted stream
+            throw new EOFException();
         }
 
         int v = _b1;
