@@ -1,16 +1,16 @@
 package org.bouncycastle.cms;
 
+import org.bouncycastle.asn1.ASN1Encodable;
+import org.bouncycastle.asn1.ASN1EncodableVector;
 import org.bouncycastle.asn1.ASN1InputStream;
 import org.bouncycastle.asn1.ASN1Null;
 import org.bouncycastle.asn1.ASN1OctetString;
-import org.bouncycastle.asn1.ASN1OutputStream;
 import org.bouncycastle.asn1.ASN1Sequence;
 import org.bouncycastle.asn1.ASN1Set;
 import org.bouncycastle.asn1.DEREncodable;
 import org.bouncycastle.asn1.DERNull;
 import org.bouncycastle.asn1.DERObject;
 import org.bouncycastle.asn1.DERObjectIdentifier;
-import org.bouncycastle.asn1.DEROutputStream;
 import org.bouncycastle.asn1.DERSet;
 import org.bouncycastle.asn1.DERTags;
 import org.bouncycastle.asn1.cms.Attribute;
@@ -23,8 +23,6 @@ import org.bouncycastle.asn1.cms.Time;
 import org.bouncycastle.asn1.x509.AlgorithmIdentifier;
 import org.bouncycastle.asn1.x509.DigestInfo;
 
-import javax.crypto.Cipher;
-import java.io.ByteArrayOutputStream;
 import java.io.IOException;
 import java.security.GeneralSecurityException;
 import java.security.InvalidKeyException;
@@ -37,6 +35,12 @@ import java.security.SignatureException;
 import java.security.cert.CertificateExpiredException;
 import java.security.cert.CertificateNotYetValidException;
 import java.security.cert.X509Certificate;
+import java.util.ArrayList;
+import java.util.Enumeration;
+import java.util.Iterator;
+import java.util.List;
+
+import javax.crypto.Cipher;
 
 /**
  * an expanded SignerInfo block from a CMS Signed message
@@ -52,14 +56,14 @@ public class SignerInformation
     private CMSProcessable          content;
     private byte[]                  signature;
     private DERObjectIdentifier     contentType;
-    private byte[]                  _digest;
-    private byte[]                  _resultDigest;
+    private DigestCalculator        digestCalculator;
+    private byte[]                  resultDigest;
 
     SignerInformation(
         SignerInfo          info,
         DERObjectIdentifier contentType,
         CMSProcessable      content,
-        byte[]              digest)
+        DigestCalculator digestCalculator)
     {
         this.info = info;
         this.sid = new SignerId();
@@ -79,12 +83,7 @@ public class SignerInformation
             {
                 IssuerAndSerialNumber   iAnds = IssuerAndSerialNumber.getInstance(s.getId());
 
-                ByteArrayOutputStream   bOut = new ByteArrayOutputStream();
-                ASN1OutputStream        aOut = new ASN1OutputStream(bOut);
-
-                aOut.writeObject(iAnds.getName());
-
-                sid.setIssuer(bOut.toByteArray());
+                sid.setIssuer(iAnds.getName().getEncoded());
                 sid.setSerialNumber(iAnds.getSerialNumber().getValue());
             }
         }
@@ -100,7 +99,7 @@ public class SignerInformation
         this.signature = info.getEncryptedDigest().getOctets();
 
         this.content = content;
-        _digest = digest;
+        this.digestCalculator = digestCalculator;
     }
 
     private byte[] encodeObj(
@@ -109,12 +108,7 @@ public class SignerInformation
     {
         if (obj != null)
         {
-            ByteArrayOutputStream   bOut = new ByteArrayOutputStream();
-            ASN1OutputStream        aOut = new ASN1OutputStream(bOut);
-
-            aOut.writeObject(obj);
-
-            return bOut.toByteArray();
+            return obj.getDERObject().getEncoded();
         }
 
         return null;
@@ -161,12 +155,12 @@ public class SignerInformation
      */
     public byte[] getContentDigest()
     {
-        if (_resultDigest == null)
+        if (resultDigest == null)
         {
             throw new IllegalStateException("method can only be called after verify.");
         }
         
-        return (byte[])_resultDigest.clone();
+        return (byte[])resultDigest.clone();
     }
     
     /**
@@ -229,6 +223,39 @@ public class SignerInformation
         return (byte[])signature.clone();
     }
 
+    /**
+     * Return a SignerInformationStore containing the counter signatures attached to this
+     * signer. If no counter signatures are present an empty store is returned.
+     */
+    public SignerInformationStore getCounterSignatures()
+    {
+        AttributeTable unsignedAttributeTable = getUnsignedAttributes();
+        if (unsignedAttributeTable == null)
+        {
+            return new SignerInformationStore(new ArrayList(0));
+        }
+        
+        List counterSignatures = new ArrayList();
+
+        Attribute counterSignatureAttribute = unsignedAttributeTable.get(CMSAttributes.counterSignature);
+        if (counterSignatureAttribute != null)
+        {
+            ASN1Set values = counterSignatureAttribute.getAttrValues();
+            counterSignatures = new ArrayList(values.size());
+
+            for (Enumeration en = values.getObjects(); en.hasMoreElements();)
+            {
+                SignerInfo si = SignerInfo.getInstance(en.nextElement());
+
+                String          digestName = CMSSignedHelper.INSTANCE.getDigestAlgName(si.getDigestAlgorithm().getObjectId().getId());
+                
+                counterSignatures.add(new SignerInformation(si, CMSAttributes.counterSignature, null, new CounterSignatureDigestCalculator(digestName, null, getSignature())));
+            }
+        }
+
+        return new SignerInformationStore(counterSignatures);
+    }
+    
     /**
      * return the DER encoding of the signed attributes.
      * @throws IOException if an encoding error occurs.
@@ -238,12 +265,7 @@ public class SignerInformation
     {
         if (signedAttributes != null)
         {
-            ByteArrayOutputStream  bOut = new ByteArrayOutputStream();
-            DEROutputStream        aOut = new DEROutputStream(bOut);
-
-            aOut.writeObject(signedAttributes);
-
-            return bOut.toByteArray();
+            return signedAttributes.getEncoded(ASN1Encodable.DER);
         }
 
         return null;
@@ -273,14 +295,14 @@ public class SignerInformation
                     content.write(
                             new CMSSignedDataGenerator.DigOutputStream(digest));
     
-                    _resultDigest = digest.digest();
+                    resultDigest = digest.digest();
                 }
                 else
                 {
-                    _resultDigest = _digest;
+                    resultDigest = digestCalculator.getDigest();
                     
                     // need to decrypt signature and check message bytes
-                    return verifyDigest(_digest, key, this.getSignature(), sigProvider);
+                    return verifyDigest(resultDigest, key, this.getSignature(), sigProvider);
                 }
             }
             else
@@ -294,12 +316,16 @@ public class SignerInformation
     
                     hash = digest.digest();
                 }
+                else if (digestCalculator != null)
+                {
+                    hash = digestCalculator.getDigest();
+                }
                 else
                 {
-                    hash = _digest;
+                    hash = null;
                 }
 
-                _resultDigest = hash;
+                resultDigest = hash;
                 
                 Attribute dig = signedAttrTable.get(
                                 CMSAttributes.messageDigest);
@@ -311,7 +337,7 @@ public class SignerInformation
                     throw new SignatureException("no hash for content found in signed attributes");
                 }
 
-                if (type == null)
+                if (type == null && !contentType.equals(CMSAttributes.counterSignature))
                 {
                     throw new SignatureException("no content type id found in signed attributes");
                 }
@@ -335,11 +361,14 @@ public class SignerInformation
                     }
                 }
 
-                DERObjectIdentifier  typeOID = (DERObjectIdentifier)type.getAttrValues().getObjectAt(0);
-
-                if (!typeOID.equals(contentType))
+                if (type != null)
                 {
-                    throw new SignatureException("contentType in signed attributes different");
+                    DERObjectIdentifier  typeOID = (DERObjectIdentifier)type.getAttrValues().getObjectAt(0);
+
+                    if (!typeOID.equals(contentType))
+                    {
+                        throw new SignatureException("contentType in signed attributes different");
+                    }
                 }
 
                 sig.update(this.getEncodedSignedAttributes());
@@ -372,7 +401,7 @@ public class SignerInformation
     
     private DigestInfo derDecode(
         byte[]  encoding)
-        throws IOException
+        throws IOException, CMSException
     {
         if (encoding[0] != (DERTags.CONSTRUCTED | DERTags.SEQUENCE))
         {
@@ -381,7 +410,16 @@ public class SignerInformation
         
         ASN1InputStream         aIn = new ASN1InputStream(encoding);
 
-        return new DigestInfo((ASN1Sequence)aIn.readObject());
+        DigestInfo digInfo = new DigestInfo((ASN1Sequence)aIn.readObject());
+
+        // length check to avoid Bleichenbacher vulnerability
+
+        if (digInfo.getEncoded().length != encoding.length)
+        {
+            throw new CMSException("malformed RSA signature");
+        }
+
+        return digInfo;
     }
     
     private boolean verifyDigest(
@@ -434,7 +472,7 @@ public class SignerInformation
                 }
                 else
                 {
-                    sig = Signature.getInstance("NONEwithDSA", sigProvider);
+                    sig = Signature.getInstance("NONEwithDSA");
                 }
                 
                 sig.initVerify(key);
@@ -448,14 +486,6 @@ public class SignerInformation
                 throw new CMSException("algorithm: " + algorithm + " not supported in base signatures.");
             }
         }
-        catch (NoSuchAlgorithmException e)
-        {
-            throw e;
-        }
-        catch (NoSuchProviderException e)
-        {
-            throw e;
-        }
         catch (GeneralSecurityException e)
         {
             throw new CMSException("Exception processing signature: " + e, e);
@@ -479,7 +509,7 @@ public class SignerInformation
     }
 
     /**
-     * verify that the given certificate succesfully handles and confirms
+     * verify that the given certificate successfully handles and confirms
      * the signature associated with this signer and, if a signingTime
      * attribute is available, that the certificate was valid at the time the
      * signature was generated.
@@ -495,14 +525,29 @@ public class SignerInformation
 
         if (attr != null)
         {
-            Attribute t = attr.get(CMSAttributes.signingTime);
-
-            if (t != null)
+            ASN1EncodableVector v = attr.getAll(CMSAttributes.signingTime);
+            switch (v.size())
             {
-                Time   time = Time.getInstance(
-                                    t.getAttrValues().getObjectAt(0).getDERObject());
+                case 0:
+                    break;
+                case 1:
+                {
+                    Attribute t = (Attribute)v.get(0);
+//                    assert t != null;
 
-                cert.checkValidity(time.getDate());
+                    ASN1Set attrValues = t.getAttrValues();
+                    if (attrValues.size() != 1)
+                    {
+                        throw new CMSException("A signing-time attribute MUST have a single attribute value");
+                    }
+
+                    Time time = Time.getInstance(attrValues.getObjectAt(0).getDERObject());
+
+                    cert.checkValidity(time.getDate());
+                    break;
+                }
+                default:
+                    throw new CMSException("The SignedAttributes in a signerInfo MUST NOT include multiple instances of the signing-time attribute");
             }
         }
 
@@ -545,5 +590,44 @@ public class SignerInformation
                     sInfo.getAuthenticatedAttributes(), sInfo.getDigestEncryptionAlgorithm(), sInfo.getEncryptedDigest(), unsignedAttr),
                     signerInformation.contentType, signerInformation.content, null);
     }
-}
 
+    /**
+     * Return a signer information object with passed in SignerInformationStore representing counter
+     * signatures attached as an unsigned attribute.
+     *
+     * @param signerInformation the signerInfo to be used as the basis.
+     * @param counterSigners signer info objects carrying counter signature.
+     * @return a copy of the original SignerInformationObject with the changed attributes.
+     */
+    public static SignerInformation addCounterSigners(
+        SignerInformation        signerInformation,
+        SignerInformationStore   counterSigners)
+    {
+        SignerInfo          sInfo = signerInformation.info;
+        AttributeTable      unsignedAttr = signerInformation.getUnsignedAttributes();
+        ASN1EncodableVector v;
+
+        if (unsignedAttr != null)
+        {
+            v = unsignedAttr.toASN1EncodableVector();
+        }
+        else
+        {
+            v = new ASN1EncodableVector();
+        }
+
+        ASN1EncodableVector sigs = new ASN1EncodableVector();
+
+        for (Iterator it = counterSigners.getSigners().iterator(); it.hasNext();)
+        {
+            sigs.add(((SignerInformation)it.next()).toSignerInfo());
+        }
+
+        v.add(new Attribute(CMSAttributes.counterSignature, new DERSet(sigs)));
+
+        return new SignerInformation(
+                new SignerInfo(sInfo.getSID(), sInfo.getDigestAlgorithm(),
+                    sInfo.getAuthenticatedAttributes(), sInfo.getDigestEncryptionAlgorithm(), sInfo.getEncryptedDigest(), new DERSet(v)),
+                    signerInformation.contentType, signerInformation.content, null);
+    }
+}
\ No newline at end of file
