@@ -1,11 +1,9 @@
 package org.bouncycastle.jce.provider;
 
-import org.bouncycastle.asn1.ASN1EncodableVector;
 import org.bouncycastle.asn1.ASN1InputStream;
 import org.bouncycastle.asn1.ASN1OctetString;
 import org.bouncycastle.asn1.ASN1OutputStream;
 import org.bouncycastle.asn1.ASN1Sequence;
-import org.bouncycastle.asn1.DEREncodable;
 import org.bouncycastle.asn1.DEREnumerated;
 import org.bouncycastle.asn1.DERGeneralizedTime;
 import org.bouncycastle.asn1.DERIA5String;
@@ -17,6 +15,7 @@ import org.bouncycastle.asn1.x509.AlgorithmIdentifier;
 import org.bouncycastle.asn1.x509.CRLDistPoint;
 import org.bouncycastle.asn1.x509.CRLNumber;
 import org.bouncycastle.asn1.x509.CRLReason;
+import org.bouncycastle.asn1.x509.CertificateList;
 import org.bouncycastle.asn1.x509.DistributionPoint;
 import org.bouncycastle.asn1.x509.DistributionPointName;
 import org.bouncycastle.asn1.x509.GeneralName;
@@ -25,12 +24,16 @@ import org.bouncycastle.asn1.x509.PolicyInformation;
 import org.bouncycastle.asn1.x509.SubjectPublicKeyInfo;
 import org.bouncycastle.asn1.x509.X509Extensions;
 import org.bouncycastle.jce.X509LDAPCertStoreParameters;
+import org.bouncycastle.jce.exception.ExtCertPathValidatorException;
 import org.bouncycastle.util.Selector;
 import org.bouncycastle.util.StoreException;
 import org.bouncycastle.x509.ExtendedPKIXParameters;
 import org.bouncycastle.x509.X509AttributeCertificate;
 import org.bouncycastle.x509.X509CRLStoreSelector;
 import org.bouncycastle.x509.X509Store;
+import org.bouncycastle.x509.ExtendedPKIXBuilderParameters;
+import org.bouncycastle.x509.X509CertStoreSelector;
+import org.bouncycastle.x509.X509AttributeCertStoreSelector;
 
 import javax.security.auth.x500.X500Principal;
 import java.io.ByteArrayOutputStream;
@@ -44,11 +47,12 @@ import java.security.cert.CertPathValidatorException;
 import java.security.cert.CertSelector;
 import java.security.cert.CertStore;
 import java.security.cert.CertStoreException;
+import java.security.cert.Certificate;
+import java.security.cert.CertificateParsingException;
 import java.security.cert.PKIXParameters;
 import java.security.cert.PolicyQualifierInfo;
 import java.security.cert.TrustAnchor;
 import java.security.cert.X509CRL;
-import java.security.cert.X509CRLEntry;
 import java.security.cert.X509CRLSelector;
 import java.security.cert.X509CertSelector;
 import java.security.cert.X509Certificate;
@@ -115,17 +119,14 @@ public class CertPathValidatorUtilities
      * @return the <code>TrustAnchor</code> object if found or
      * <code>null</code> if not.
      *
-     * @exception CertPathValidatorException if a TrustAnchor  was
-     * found but the signature verification on the given certificate
-     * has thrown an exception. This Exception can be obtainted with
-     * <code>getCause()</code> method.
-     **/
-    protected static final TrustAnchor findTrustAnchor(
+     * @exception AnnotatedException
+     *                if a TrustAnchor was found but the signature verification
+     *                on the given certificate has thrown an exception.
+     */
+    protected static TrustAnchor findTrustAnchor(
         X509Certificate cert,
-        CertPath        certPath,
-        int             index,
         Set             trustAnchors) 
-        throws CertPathValidatorException
+            throws AnnotatedException
     {
         Iterator iter = trustAnchors.iterator();
         TrustAnchor trust = null;
@@ -140,7 +141,7 @@ public class CertPathValidatorUtilities
         }
         catch (IOException ex)
         {
-            throw new CertPathValidatorException(ex);
+            throw new AnnotatedException("Cannot set subject search criteria for trust anchor.", ex);
         }
 
         while (iter.hasNext() && trust == null)
@@ -199,12 +200,35 @@ public class CertPathValidatorUtilities
 
         if (trust == null && invalidKeyEx != null)
         {
-            throw new CertPathValidatorException("TrustAnchor found but certificate validation failed.", invalidKeyEx, certPath, index);
+            throw new AnnotatedException("TrustAnchor found but certificate validation failed.", invalidKeyEx);
         }
 
         return trust;
     }
 
+    protected static void addAdditionalStoresFromAltNames(
+            X509Certificate cert,
+            ExtendedPKIXParameters pkixParams)
+            throws CertificateParsingException
+    {
+        // if in the IssuerAltName extension an URI
+        // is given, add an additinal X.509 store
+        if (cert.getIssuerAlternativeNames() != null)
+        {
+            Iterator it = cert.getIssuerAlternativeNames().iterator();
+            while (it.hasNext())
+            {
+                // look for URI
+                List list = (List) it.next();
+                if (list.get(0).equals(new Integer(GeneralName.uniformResourceIdentifier)))
+                {
+                    // found
+                    String temp = (String) list.get(1);
+                    CertPathValidatorUtilities.addAdditionalStoreFromLocation(temp, pkixParams);
+                }
+            }
+        }
+    }
     /**
      * Returns the issuer of an attribute certificate or certificate.
      * @param cert The attribute certificate or certificate.
@@ -247,7 +271,14 @@ public class CertPathValidatorUtilities
     
     
     /**
-     * extract the value of the given extension, if it exists.
+     * Extract the value of the given extension, if it exists.
+     * 
+     * @param ext
+     *            The extension object.
+     * @param oid
+     *            The object identifier to obtain.
+     * @throws AnnotatedException
+     *             if the extension cannot be read.
      */
     protected static DERObject getExtensionValue(
         java.security.cert.X509Extension    ext,
@@ -276,7 +307,7 @@ public class CertPathValidatorUtilities
             aIn = new ASN1InputStream(octs.getOctets());
             return aIn.readObject();
         }
-        catch (IOException e)
+        catch (Exception e)
         {
             throw new AnnotatedException("exception processing extension " + oid, e);
         }
@@ -299,52 +330,14 @@ public class CertPathValidatorUtilities
 
             return info.getAlgorithmId();
         }
-        catch (IOException e)
+        catch (Exception e)
         {
-            throw new CertPathValidatorException("exception processing public key");
+            throw new ExtCertPathValidatorException("Subject public key cannot be decoded.", e);
         }
     }
     
     // crl checking
-    
-    /**
-     * Return a Collection of all CRLs found in the
-     * CertStore's that are matching the crlSelect criteriums.
-     *
-     * @param crlSelect a {@link CertSelector CertSelector}
-     * object that will be used to select the CRLs
-     * @param crlStores a List containing only {@link CertStore
-     * CertStore} objects. These are used to search for
-     * CRLs
-     *
-     * @return a Collection of all found {@link CRL CRL}
-     * objects. May be empty but never <code>null</code>.
-     */
-    protected static final Collection findCRLs(
-        X509CRLSelector crlSelect,
-        List            crlStores)
-        throws AnnotatedException
-    {
-        Set crls = new HashSet();
-        Iterator iter = crlStores.iterator();
-    
-        while (iter.hasNext())
-        {
-            CertStore   certStore = (CertStore)iter.next();
-    
-            try
-            {
-                crls.addAll(certStore.getCRLs(crlSelect));
-            }
-            catch (CertStoreException e)
-            {
-                throw new AnnotatedException("cannot extract crl: " + e, e);
-            }
-        }
-    
-        return crls;
-    }
-    
+
     /**
      * Return a Collection of all CRLs found in the X509Store's that are
      * matching the crlSelect criteriums.
@@ -369,17 +362,37 @@ public class CertPathValidatorUtilities
 
         while (iter.hasNext())
         {
-            X509Store store = (X509Store) iter.next();
+            Object obj = iter.next();
 
-            try
+            if (obj instanceof X509Store)
             {
-                crls.addAll(store.getMatches(crlSelect));
-                foundValidStore = true;
+                X509Store store = (X509Store)obj;
+
+                try
+                {
+                    crls.addAll(store.getMatches(crlSelect));
+                    foundValidStore = true;
+                }
+                catch (StoreException e)
+                {
+                    lastException = new AnnotatedException(
+                        "Exception searching in X.509 CRL store.", e);
+                }
             }
-            catch (StoreException e)
+            else
             {
-                lastException = new AnnotatedException(
-                    "Exception searching in X.509 CRL store.", e);
+                CertStore store = (CertStore)obj;
+
+                try
+                {
+                    crls.addAll(store.getCRLs(crlSelect));
+                    foundValidStore = true;
+                }
+                catch (CertStoreException e)
+                {
+                    lastException = new AnnotatedException(
+                        "Exception searching in X.509 CRL store.", e);
+                }
             }
         }
         if (!foundValidStore && lastException != null)
@@ -418,7 +431,7 @@ public class CertPathValidatorUtilities
             }
             catch (IOException ex)
             {
-                throw new CertPathValidatorException("exception building qualifier set: " + ex);
+                throw new ExtCertPathValidatorException("Policy qualifier info cannot be decoded.", ex);
             }
     
             bOut.reset();
@@ -572,14 +585,41 @@ public class CertPathValidatorUtilities
                 if (ANY_POLICY.equals(node.getValidPolicy()))
                 {
                     Set pq = null;
-                    ASN1Sequence policies = (ASN1Sequence)getExtensionValue(cert, CERTIFICATE_POLICIES);
+                    ASN1Sequence policies = null;
+                    try
+                    {
+                        policies = DERSequence.getInstance(getExtensionValue(cert, CERTIFICATE_POLICIES));
+                    }
+                    catch (Exception e)
+                    {
+                        throw
+
+                        new AnnotatedException("Certificate policies cannot be decoded.", e);
+                    }
                     Enumeration e = policies.getObjects();
                     while (e.hasMoreElements())
                     {
-                        PolicyInformation pinfo = PolicyInformation.getInstance(e.nextElement());
+                        PolicyInformation pinfo = null;
+
+                        try
+                        {
+                            pinfo = PolicyInformation.getInstance(e.nextElement());
+                        }
+                        catch (Exception ex)
+                        {
+                            throw new AnnotatedException("Policy information cannot be decoded.", ex);
+                        }
                         if (ANY_POLICY.equals(pinfo.getPolicyIdentifier().getId()))
                         {
+                            try
+                            {
                             pq = getQualifierSet(pinfo.getPolicyQualifiers());
+                            }
+                            catch (CertPathValidatorException ex)
+                            {
+                                throw new ExtCertPathValidatorException(
+                                        "Policy qualifier info set could not be built.", ex);
+                            }
                             break;
                         }
                     }
@@ -695,45 +735,6 @@ public class CertPathValidatorUtilities
         }
     }
 
-    /**
-     * Return a Collection of all certificates found in the CertStore's that are
-     * matching the certSelect criteriums.
-     *
-     * @param certSelect a {@link CertSelector CertSelector} object that will
-     *            be used to select the certificates
-     * @param certStores a List containing only {@link CertStore CertStore}
-     *            objects. These are used to search for certificates
-     *
-     * @return a Collection of all found {@link java.security.cert.Certificate Certificate}
-     *         objects. May be empty but never <code>null</code>.
-     */
-    protected static Collection findCertificates(CertSelector certSelect,
-        List certStores) throws AnnotatedException
-    {
-        Set certs = new HashSet();
-        Iterator iter = certStores.iterator();
-
-        while (iter.hasNext())
-        {
-            CertStore certStore = (CertStore) iter.next();
-
-            try
-            {
-                certs.addAll(certStore.getCertificates(certSelect));
-            }
-            catch (CertStoreException e)
-            {
-                throw
-
-                new AnnotatedException(
-                    "Problem while picking certificates from certificate store.",
-                    e);
-            }
-        }
-
-        return certs;
-    }
-
     /**
      * Return a Collection of all certificates or attribute certificates found
      * in the X509Store's that are matching the certSelect criteriums.
@@ -747,7 +748,7 @@ public class CertPathValidatorUtilities
      *         {@link org.bouncycastle.x509.X509AttributeCertificate} objects.
      *         May be empty but never <code>null</code>.
      */
-    protected static Collection findCertificates(Selector certSelect,
+    protected static Collection findCertificates(X509CertStoreSelector certSelect,
         List certStores) throws AnnotatedException
     {
         Set certs = new HashSet();
@@ -755,17 +756,67 @@ public class CertPathValidatorUtilities
 
         while (iter.hasNext())
         {
-            X509Store certStore = (X509Store) iter.next();
-            try
-            {
-                certs.addAll(certStore.getMatches(certSelect));
-            }
-            catch (StoreException e)
-            {
-                throw
+            Object obj = iter.next();
 
-                new AnnotatedException(
-                    "Problem while picking certificates from X.509 store.", e);
+            if (obj instanceof X509Store)
+            {
+                X509Store certStore = (X509Store)obj;
+                try
+                {
+                    certs.addAll(certStore.getMatches(certSelect));
+                }
+                catch (StoreException e)
+                {
+                    throw
+
+                    new AnnotatedException(
+                        "Problem while picking certificates from X.509 store.", e);
+                }
+            }
+            else
+            {
+                CertStore certStore = (CertStore)obj;
+
+                try
+                {
+                    certs.addAll(certStore.getCertificates(certSelect));
+                }
+                catch (CertStoreException e)
+                {
+                    throw new AnnotatedException(
+                        "Problem while picking certificates from certificate store.",
+                        e);
+                }
+            }
+        }
+        return certs;
+    }
+
+    protected static Collection findCertificates(X509AttributeCertStoreSelector certSelect,
+                                                 List certStores)
+    throws AnnotatedException
+    {
+        Set certs = new HashSet();
+        Iterator iter = certStores.iterator();
+
+        while (iter.hasNext())
+        {
+            Object obj = iter.next();
+
+            if (obj instanceof X509Store)
+            {
+                X509Store certStore = (X509Store)obj;
+                try
+                {
+                    certs.addAll(certStore.getMatches(certSelect));
+                }
+                catch (StoreException e)
+                {
+                    throw
+
+                        new AnnotatedException(
+                            "Problem while picking certificates from X.509 store.", e);
+                }
             }
         }
         return certs;
@@ -791,20 +842,23 @@ public class CertPathValidatorUtilities
             {
                 DistributionPointName dpn = dps[i].getDistributionPoint();
                 // look for URIs in fullName
-                if (dpn.getType() == DistributionPointName.FULL_NAME)
+                if (dpn != null)
                 {
-                    GeneralName[] genNames = GeneralNames.getInstance(
-                        dpn.getName()).getNames();
-                    // look for an URI
-                    for (int j = 0; j < genNames.length; j++)
+                    if (dpn.getType() == DistributionPointName.FULL_NAME)
                     {
-                        if (genNames[j].getTagNo() == GeneralName.uniformResourceIdentifier)
+                        GeneralName[] genNames = GeneralNames.getInstance(
+                            dpn.getName()).getNames();
+                        // look for an URI
+                        for (int j = 0; j < genNames.length; j++)
                         {
-                            String location = DERIA5String.getInstance(
-                                genNames[j].getName()).getString();
-                            CertPathValidatorUtilities
-                                .addAdditionalStoreFromLocation(location,
-                                    pkixParams);
+                            if (genNames[j].getTagNo() == GeneralName.uniformResourceIdentifier)
+                            {
+                                String location = DERIA5String.getInstance(
+                                    genNames[j].getName()).getString();
+                                CertPathValidatorUtilities
+                                    .addAdditionalStoreFromLocation(location,
+                                        pkixParams);
+                            }
                         }
                     }
                 }
@@ -823,7 +877,7 @@ public class CertPathValidatorUtilities
      * <code>X500Principal</code>.
      *
      * @param dp The distribution point.
-     * @param issuerPrincipals The issuers of the certificate or atribute
+     * @param issuerPrincipals The issuers of the certificate or attribute
      *            certificate which contains the distribution point.
      * @param selector The CRL selector.
      * @param pkixParams The PKIX parameters containing the cert stores.
@@ -832,8 +886,10 @@ public class CertPathValidatorUtilities
      * contain only <code>X500Principal</code>s.
      */
     protected static void getCRLIssuersFromDistributionPoint(
-        DistributionPoint dp, Collection issuerPrincipals,
-        X509CRLStoreSelector selector, ExtendedPKIXParameters pkixParams)
+        DistributionPoint dp,
+        Collection issuerPrincipals,
+        X509CRLSelector selector,
+        ExtendedPKIXParameters pkixParams)
         throws AnnotatedException
     {
         List issuers = new ArrayList();
@@ -877,49 +933,50 @@ public class CertPathValidatorUtilities
                 issuers.add((X500Principal)it.next());
             }
         }
+        // TODO: is not found although this should correctly add the rel name. selector of Sun is buggy here or PKI test case is invalid
         // distributionPoint
-        if (dp.getDistributionPoint() != null)
-        {
-            // look for nameRelativeToCRLIssuer
-            if (dp.getDistributionPoint().getType() == DistributionPointName.NAME_RELATIVE_TO_CRL_ISSUER)
-            {
-                // append fragment to issuer, only one
-                // issuer can be there, if this is given
-                if (issuers.size() != 1)
-                {
-                    throw new AnnotatedException(
-                        "nameRelativeToCRLIssuer field is given but more than one CRL issuer is given.");
-                }
-                DEREncodable relName = dp.getDistributionPoint().getName();
-                Iterator it = issuers.iterator();
-                List issuersTemp = new ArrayList(issuers.size());
-                while (it.hasNext())
-                {
-                    Enumeration e = null;
-                    try
-                    {
-                        e = ASN1Sequence.getInstance(
-                            new ASN1InputStream(((X500Principal) it.next())
-                                .getEncoded()).readObject()).getObjects();
-                    }
-                    catch (IOException ex)
-                    {
-                        throw new AnnotatedException(
-                            "Cannot decode CRL issuer information.", ex);
-                    }
-                    ASN1EncodableVector v = new ASN1EncodableVector();
-                    while (e.hasMoreElements())
-                    {
-                        v.add((DEREncodable) e.nextElement());
-                    }
-                    v.add(relName);
-                    issuersTemp.add(new X500Principal(new DERSequence(v)
-                        .getDEREncoded()));
-                }
-                issuers.clear();
-                issuers.addAll(issuersTemp);
-            }
-        }
+//        if (dp.getDistributionPoint() != null)
+//        {
+//            // look for nameRelativeToCRLIssuer
+//            if (dp.getDistributionPoint().getType() == DistributionPointName.NAME_RELATIVE_TO_CRL_ISSUER)
+//            {
+//                // append fragment to issuer, only one
+//                // issuer can be there, if this is given
+//                if (issuers.size() != 1)
+//                {
+//                    throw new AnnotatedException(
+//                        "nameRelativeToCRLIssuer field is given but more than one CRL issuer is given.");
+//                }
+//                DEREncodable relName = dp.getDistributionPoint().getName();
+//                Iterator it = issuers.iterator();
+//                List issuersTemp = new ArrayList(issuers.size());
+//                while (it.hasNext())
+//                {
+//                    Enumeration e = null;
+//                    try
+//                    {
+//                        e = ASN1Sequence.getInstance(
+//                            new ASN1InputStream(((X500Principal) it.next())
+//                                .getEncoded()).readObject()).getObjects();
+//                    }
+//                    catch (IOException ex)
+//                    {
+//                        throw new AnnotatedException(
+//                            "Cannot decode CRL issuer information.", ex);
+//                    }
+//                    ASN1EncodableVector v = new ASN1EncodableVector();
+//                    while (e.hasMoreElements())
+//                    {
+//                        v.add((DEREncodable) e.nextElement());
+//                    }
+//                    v.add(relName);
+//                    issuersTemp.add(new X500Principal(new DERSequence(v)
+//                        .getDEREncoded()));
+//                }
+//                issuers.clear();
+//                issuers.addAll(issuersTemp);
+//            }
+//        }
         Iterator it = issuers.iterator();
         while (it.hasNext())
         {
@@ -935,15 +992,42 @@ public class CertPathValidatorUtilities
         }
     }
 
-    protected static void getCertStatus(Date validDate, X509CRL crl,
-        BigInteger serialNumber, CertStatus certStatus)
+    private static BigInteger getSerialNumber(
+            Object cert)
+    {
+        if (cert instanceof X509Certificate)
+        {
+            return ((X509Certificate) cert).getSerialNumber();
+        }
+        else
+        {
+            return ((X509AttributeCertificate) cert).getSerialNumber();
+        }
+    }
+    
+    protected static void getCertStatus(
+            Date validDate,
+            X509CRL crl,
+            Object cert,
+            CertStatus certStatus)
         throws AnnotatedException
     {
-        // (i) or (j)
-        // TODO: If two certificates from different issuers in indirect CRLs have
-        // the same serial number ...
-        X509CRLEntry crl_entry = crl.getRevokedCertificate(serialNumber);
-        if (crl_entry != null)
+        // use BC X509CRLObject so that indirect CRLs are supported
+        X509CRLObject bcCRL = null;
+        try
+        {
+            bcCRL = new X509CRLObject(new CertificateList((ASN1Sequence) ASN1Sequence.fromByteArray(crl.getEncoded())));
+        }
+        catch (Exception exception)
+        {
+            throw new AnnotatedException("Bouncy Castle X509CRLObject could not be created.", exception);
+        }
+        // use BC X509CRLEntryObject, so that getCertificateIssuer() is
+        // supported.
+        X509CRLEntryObject crl_entry = (X509CRLEntryObject) bcCRL.getRevokedCertificate(getSerialNumber(cert));
+        if (crl_entry != null
+                && (getEncodedIssuerPrincipal(cert).equals(crl_entry.getCertificateIssuer()) || getEncodedIssuerPrincipal(cert)
+                        .equals(getIssuerPrincipal(crl))))
         {
             DEREnumerated reasonCode = null;
             if (crl_entry.hasExtensions())
@@ -996,17 +1080,23 @@ public class CertPathValidatorUtilities
      * @param completeCRL The complete CRL the delta CRL is for.
      * @return A <code>Set</code> of <code>X509CRL</code>s with delta CRLs.
      * @throws AnnotatedException if an exception occurs while picking the delta
-     *             CRLs or no delta CRLs are found.
+     *             CRLs.
      */
     protected static Set getDeltaCRLs(Date currentDate,
         ExtendedPKIXParameters paramsPKIX, X509CRL completeCRL)
         throws AnnotatedException
     {
 
-        Set set = new HashSet();
-
         X509CRLStoreSelector deltaSelect = new X509CRLStoreSelector();
-        deltaSelect.setDateAndTime(currentDate);
+
+        if (paramsPKIX.getDate() != null)
+        {
+            deltaSelect.setDateAndTime(paramsPKIX.getDate());
+        }
+        else
+        {
+            deltaSelect.setDateAndTime(currentDate);
+        }
 
         // 5.2.4 (a)
         try
@@ -1019,12 +1109,15 @@ public class CertPathValidatorUtilities
             new AnnotatedException("Cannot extract issuer from CRL.", e);
         }
 
-        BigInteger completeCRLNumber;
+        BigInteger completeCRLNumber = null;
         try
         {
-            completeCRLNumber = CRLNumber.getInstance(
-                CertPathValidatorUtilities.getExtensionValue(completeCRL,
-                    CRL_NUMBER)).getPositiveValue();
+            DERObject derObect = CertPathValidatorUtilities.getExtensionValue(completeCRL,
+                    CRL_NUMBER);
+            if (derObect != null)
+            {
+                completeCRLNumber = CRLNumber.getInstance(derObect).getPositiveValue();
+            }
         }
         catch (Exception e)
         {
@@ -1044,35 +1137,31 @@ public class CertPathValidatorUtilities
                 "Issuing distribution point extension value could not be read.",
                 e);
         }
+
+        // 5.2.4 (d)
+
+        deltaSelect.setMinCRLNumber(completeCRLNumber == null ? null : completeCRLNumber
+            .add(BigInteger.valueOf(1)));
+
         deltaSelect.setIssuingDistributionPoint(idp);
         deltaSelect.setIssuingDistributionPointEnabled(true);
 
         // 5.2.4 (c)
         deltaSelect.setMaxBaseCRLNumber(completeCRLNumber);
 
-        // 5.2.4 (d)
-
-        deltaSelect.setMinCRLNumber(completeCRLNumber
-            .add(BigInteger.valueOf(1)));
-
         Set temp = new HashSet();
         // find delta CRLs
         try
         {
-            temp.addAll(CertPathValidatorUtilities.findCRLs(deltaSelect,
-                paramsPKIX.getAddionalStores()));
-            temp.addAll(CertPathValidatorUtilities.findCRLs(deltaSelect,
-                paramsPKIX.getStores()));
+            temp.addAll(CertPathValidatorUtilities.findCRLs(deltaSelect, paramsPKIX.getAdditionalStores()));
+            temp.addAll(CertPathValidatorUtilities.findCRLs(deltaSelect, paramsPKIX.getStores()));
+            temp.addAll(CertPathValidatorUtilities.findCRLs(deltaSelect, paramsPKIX.getCertStores()));
         }
         catch (AnnotatedException e)
         {
             throw new AnnotatedException("Could not search for delta CRLs.", e);
         }
-        if (set.isEmpty())
-        {
-            throw new AnnotatedException("No delta CRLs found.");
-        }
-        return set;
+        return temp;
     }
 
     /**
@@ -1104,10 +1193,9 @@ public class CertPathValidatorUtilities
             }
             else
             {
-                issuers.add(((X509Certificate)cert).getSubjectX500Principal());
+                issuers.add(getEncodedIssuerPrincipal(cert));
             }
-            CertPathValidatorUtilities.getCRLIssuersFromDistributionPoint(dp,
-                issuers, crlselect, paramsPKIX);
+            CertPathValidatorUtilities.getCRLIssuersFromDistributionPoint(dp, issuers, crlselect, paramsPKIX);
         }
         catch (AnnotatedException e)
         {
@@ -1116,23 +1204,30 @@ public class CertPathValidatorUtilities
         }
         if (cert instanceof X509Certificate)
         {
-            crlselect.setCertificateChecking((X509Certificate) cert);
+            crlselect.setCertificateChecking((X509Certificate)cert);
+        }
+        else if (cert instanceof X509AttributeCertificate)
+        {
+            crlselect.setAttrCertificateChecking((X509AttributeCertificate)cert);
+        }
+
+        if (paramsPKIX.getDate() != null)
+        {
+            crlselect.setDateAndTime(paramsPKIX.getDate());
         }
         else
         {
-            crlselect
-                .setAttrCertificateChecking((X509AttributeCertificate) cert);
+            crlselect.setDateAndTime(currentDate);
         }
-        crlselect.setDateAndTime(currentDate);
+
         crlselect.setCompleteCRLEnabled(true);
 
         Set crls = new HashSet();
         try
         {
-            crls.addAll(CertPathValidatorUtilities.findCRLs(crlselect,
-                paramsPKIX.getStores()));
-            crls.addAll(CertPathValidatorUtilities.findCRLs(crlselect,
-                paramsPKIX.getAddionalStores()));
+            crls.addAll(CertPathValidatorUtilities.findCRLs(crlselect, paramsPKIX.getStores()));
+            crls.addAll(CertPathValidatorUtilities.findCRLs(crlselect, paramsPKIX.getAdditionalStores()));
+            crls.addAll(CertPathValidatorUtilities.findCRLs(crlselect, paramsPKIX.getCertStores()));
         }
         catch (AnnotatedException e)
         {
@@ -1225,9 +1320,10 @@ public class CertPathValidatorUtilities
      *         <code>index</code> extended with DSA parameters if applicable.
      * @throws AnnotatedException if DSA parameters cannot be inherited.
      */
-    protected static PublicKey getNextWorkingKey(X509Certificate cert, List certs, int index)
+    protected static PublicKey getNextWorkingKey(List certs, int index)
         throws CertPathValidatorException
     {
+        Certificate cert = (Certificate) certs.get(index);
         PublicKey pubKey = cert.getPublicKey();
         if (!(pubKey instanceof DSAPublicKey))
         {
@@ -1262,10 +1358,67 @@ public class CertPathValidatorUtilities
             }
             catch (Exception exception)
             {
-                throw new CertPathValidatorException(exception);
+                throw new RuntimeException(exception.getMessage());
             }
         }
-        throw new CertPathValidatorException(
-            "DSA parameters cannot be inherited from previous certificate.");
+        throw new CertPathValidatorException("DSA parameters cannot be inherited from previous certificate.");
     }
+    
+    /**
+     * Find the issuer certificates of a given certificate.
+     * 
+     * @param cert
+     *            The certificate for which an issuer should be found.
+     * @param pkixParams
+     * @return A <code>Collection</code> object containing the issuer
+     *         <code>X509Certificate</code>s. Never <code>null</code>.
+     * 
+     * @exception AnnotatedException
+     *                if an error occurs.
+     */
+    protected static Collection findIssuerCerts(
+        X509Certificate cert,
+        ExtendedPKIXBuilderParameters pkixParams)
+            throws AnnotatedException
+    {
+        X509CertStoreSelector certSelect = new X509CertStoreSelector();
+        Set certs = new HashSet();
+        try
+        {
+            certSelect.setSubject(cert.getIssuerX500Principal().getEncoded());
+        }
+        catch (IOException ex)
+        {
+            throw new AnnotatedException(
+                    "Subject criteria for certificate selector to find issuer certificate could not be set.", ex);
+        }
+
+        Iterator iter;
+
+        try
+        {
+            List matches = new ArrayList();
+
+            matches.addAll(CertPathValidatorUtilities.findCertificates(certSelect, pkixParams.getCertStores()));
+            matches.addAll(CertPathValidatorUtilities.findCertificates(certSelect, pkixParams.getStores()));
+            matches.addAll(CertPathValidatorUtilities.findCertificates(certSelect, pkixParams.getAdditionalStores()));
+
+            iter = matches.iterator();
+        }
+        catch (AnnotatedException e)
+        {
+            throw new AnnotatedException("Issuer certificate cannot be searched.", e);
+        }
+
+        X509Certificate issuer = null;
+        while (iter.hasNext())
+        {
+            issuer = (X509Certificate) iter.next();
+            // issuer cannot be verified because possible DSA inheritance
+            // parameters are missing
+            certs.add(issuer);
+        }
+        return certs;
+    }
+    
 }
