@@ -12,6 +12,7 @@ import org.bouncycastle.asn1.DERObjectIdentifier;
 import org.bouncycastle.asn1.x509.AccessDescription;
 import org.bouncycastle.asn1.x509.AlgorithmIdentifier;
 import org.bouncycastle.asn1.x509.AuthorityInformationAccess;
+import org.bouncycastle.asn1.x509.AuthorityKeyIdentifier;
 import org.bouncycastle.asn1.x509.BasicConstraints;
 import org.bouncycastle.asn1.x509.CRLDistPoint;
 import org.bouncycastle.asn1.x509.DistributionPoint;
@@ -33,8 +34,10 @@ import org.bouncycastle.i18n.filter.UntrustedInput;
 import org.bouncycastle.i18n.filter.UntrustedUrlInput;
 import org.bouncycastle.jce.provider.AnnotatedException;
 import org.bouncycastle.jce.provider.CertPathValidatorUtilities;
-import org.bouncycastle.jce.provider.PKIXNameConstraints;
+import org.bouncycastle.jce.provider.PKIXNameConstraintValidator;
 import org.bouncycastle.jce.provider.PKIXPolicyNode;
+import org.bouncycastle.jce.provider.PKIXNameConstraintValidatorException;
+import org.bouncycastle.x509.extension.X509ExtensionUtil;
 
 import javax.security.auth.x500.X500Principal;
 import java.io.ByteArrayInputStream;
@@ -383,7 +386,7 @@ public class PKIXCertPathReviewer extends CertPathValidatorUtilities
         //
         
         // (b)  and (c)
-        PKIXNameConstraints nameConstraints = new PKIXNameConstraints();
+        PKIXNameConstraintValidator nameConstraintValidator = new PKIXNameConstraintValidator();
 
         //
         // process each certificate except the last in the path
@@ -424,9 +427,9 @@ public class PKIXCertPathReviewer extends CertPathValidatorUtilities
     
                     try
                     {
-                        nameConstraints.checkPermittedDN(dns);
+                        nameConstraintValidator.checkPermittedDN(dns);
                     }
-                    catch (CertPathValidatorException cpve)
+                    catch (PKIXNameConstraintValidatorException cpve)
                     {
                         ErrorBundle msg = new ErrorBundle(RESOURCE_NAME,"CertPathReviewer.notPermittedDN", 
                                 new Object[] {new UntrustedInput(principal.getName())});
@@ -435,9 +438,9 @@ public class PKIXCertPathReviewer extends CertPathValidatorUtilities
                     
                     try
                     {
-                        nameConstraints.checkExcludedDN(dns);
+                        nameConstraintValidator.checkExcludedDN(dns);
                     }
-                    catch (CertPathValidatorException cpve)
+                    catch (PKIXNameConstraintValidatorException cpve)
                     {
                         ErrorBundle msg = new ErrorBundle(RESOURCE_NAME,"CertPathReviewer.excludedDN",
                                 new Object[] {new UntrustedInput(principal.getName())});
@@ -463,10 +466,10 @@ public class PKIXCertPathReviewer extends CertPathValidatorUtilities
 
                             try
                             {
-                                nameConstraints.checkPermitted(name);
-                                nameConstraints.checkExcluded(name);
+                                nameConstraintValidator.checkPermitted(name);
+                                nameConstraintValidator.checkExcluded(name);
                             }
-                            catch (CertPathValidatorException cpve)
+                            catch (PKIXNameConstraintValidatorException cpve)
                             {
                                 ErrorBundle msg = new ErrorBundle(RESOURCE_NAME,"CertPathReviewer.notPermittedEmail",
                                         new Object[] {new UntrustedInput(name)});
@@ -585,13 +588,7 @@ public class PKIXCertPathReviewer extends CertPathValidatorUtilities
                     ASN1Sequence permitted = nc.getPermittedSubtrees();
                     if (permitted != null)
                     {
-                        Enumeration e = permitted.getObjects();
-                        while (e.hasMoreElements())
-                        {
-                            GeneralSubtree  subtree = GeneralSubtree.getInstance(e.nextElement());
-
-                            nameConstraints.intersectPermittedSubtree(subtree);
-                        }
+                        nameConstraintValidator.intersectPermittedSubtree(permitted);
                     }
                 
                     //
@@ -605,7 +602,7 @@ public class PKIXCertPathReviewer extends CertPathValidatorUtilities
                         {
                             GeneralSubtree  subtree = GeneralSubtree.getInstance(e.nextElement());
 
-                            nameConstraints.addExcludedSubtree(subtree);
+                            nameConstraintValidator.addExcludedSubtree(subtree);
                         }
                     }
                 }
@@ -869,9 +866,50 @@ public class PKIXCertPathReviewer extends CertPathValidatorUtilities
                     addError(msg,index);
                 }
             }
+            else if (isSelfIssued(cert))
+            {
+                try
+                {
+                    cert.verify(cert.getPublicKey(), "BC");
+                    ErrorBundle msg = new ErrorBundle(RESOURCE_NAME,"CertPathReviewer.rootKeyIsValidButNotATrustAnchor");
+                    addError(msg, index);
+                }
+                catch (GeneralSecurityException ex)
+                {
+                    ErrorBundle msg = new ErrorBundle(RESOURCE_NAME,"CertPathReviewer.signatureNotVerified",
+                            new Object[] {ex.getMessage(),ex,ex.getClass().getName()}); 
+                    addError(msg,index);
+                }
+            }
             else
             {
                 ErrorBundle msg = new ErrorBundle(RESOURCE_NAME,"CertPathReviewer.NoIssuerPublicKey");
+                // if there is an authority key extension add the serial and issuer of the missing certificate
+                byte[] akiBytes = cert.getExtensionValue(X509Extensions.AuthorityKeyIdentifier.getId());
+                if (akiBytes != null)
+                {
+                    try
+                    {
+                        AuthorityKeyIdentifier aki = AuthorityKeyIdentifier.getInstance(
+                            X509ExtensionUtil.fromExtensionValue(akiBytes));
+                        GeneralNames issuerNames = aki.getAuthorityCertIssuer();
+                        if (issuerNames != null)
+                        {
+                            GeneralName name = issuerNames.getNames()[0];
+                            BigInteger serial = aki.getAuthorityCertSerialNumber(); 
+                            if (serial != null)
+                            {
+                                Object[] extraArgs = {new LocaleString(RESOURCE_NAME, "missingIssuer"), " \"", name , 
+                                        "\" ", new LocaleString(RESOURCE_NAME, "missingSerial") , " ", serial};
+                                msg.setExtraArguments(extraArgs);
+                            }
+                        }
+                    }
+                    catch (IOException e)
+                    {
+                        // ignore
+                    }
+                }
                 addError(msg,index);
             }
 
@@ -1034,7 +1072,7 @@ public class PKIXCertPathReviewer extends CertPathValidatorUtilities
 
             try
             {
-                workingPublicKey = getNextWorkingKey(sign, certs, index);
+                workingPublicKey = getNextWorkingKey(certs, index);
                 workingAlgId = getAlgorithmIdentifier(workingPublicKey);
                 workingPublicKeyAlgorithm = workingAlgId.getObjectId();
                 workingPublicKeyParameters = workingAlgId.getParameters();
@@ -1987,8 +2025,8 @@ public class PKIXCertPathReviewer extends CertPathValidatorUtilities
             int index) 
         throws CertPathReviewerException
     {
-        X509CRLSelector crlselect;
-        crlselect = new X509CRLSelector();
+        X509CRLStoreSelector crlselect;
+        crlselect = new X509CRLStoreSelector();
         
         try
         {
@@ -2011,7 +2049,7 @@ public class PKIXCertPathReviewer extends CertPathValidatorUtilities
             if (crl_coll.isEmpty())
             {
                 // notifcation - no local crls found
-                crl_coll = findCRLs(new X509CRLSelector(),paramsPKIX.getCertStores());
+                crl_coll = findCRLs(new X509CRLStoreSelector(),paramsPKIX.getCertStores());
                 Iterator it = crl_coll.iterator();
                 List nonMatchingCrlNames = new ArrayList();
                 while (it.hasNext())
@@ -2236,7 +2274,7 @@ public class PKIXCertPathReviewer extends CertPathValidatorUtilities
 
             if (dci != null)
             {
-                X509CRLSelector baseSelect = new X509CRLSelector();
+                X509CRLStoreSelector baseSelect = new X509CRLStoreSelector();
 
                 try
                 {
